@@ -1,6 +1,6 @@
 package Zonemaster::Engine::Test::Nameserver;
 
-use version; our $VERSION = version->declare("v1.0.9");
+use version; our $VERSION = version->declare("v1.0.23");
 
 use strict;
 use warnings;
@@ -13,6 +13,13 @@ use Zonemaster::Engine::Test::Address;
 use Zonemaster::Engine::Constants qw[:ip];
 
 use List::MoreUtils qw[uniq none];
+use Readonly;
+
+Readonly my @NONEXISTENT_NAMES => qw{
+  xn--nameservertest.iis.se
+  xn--nameservertest.icann.org
+  xn--nameservertest.ripe.net
+};
 
 ###
 ### Entry Points
@@ -22,33 +29,45 @@ sub all {
     my ( $class, $zone ) = @_;
     my @results;
 
-    if ( Zonemaster::Engine->config->should_run( 'nameserver01' ) ) {
+    if ( Zonemaster::Engine::Util::should_run_test( q{nameserver01} ) ) {
         push @results, $class->nameserver01( $zone );
     }
-    if ( Zonemaster::Engine->config->should_run( 'nameserver02' ) ) {
+    if ( Zonemaster::Engine::Util::should_run_test( q{nameserver02} ) ) {
         push @results, $class->nameserver02( $zone );
     }
-    if ( Zonemaster::Engine->config->should_run( 'nameserver03' ) ) {
+    if ( Zonemaster::Engine::Util::should_run_test( q{nameserver03} ) ) {
         push @results, $class->nameserver03( $zone );
     }
-    if ( Zonemaster::Engine->config->should_run( 'nameserver04' ) ) {
+    if ( Zonemaster::Engine::Util::should_run_test( q{nameserver04} ) ) {
         push @results, $class->nameserver04( $zone );
     }
-    if ( Zonemaster::Engine->config->should_run( 'nameserver05' ) ) {
+    if ( Zonemaster::Engine::Util::should_run_test( q{nameserver05} ) ) {
         push @results, $class->nameserver05( $zone );
     }
-    if ( Zonemaster::Engine->config->should_run( 'nameserver06' ) ) {
+    if ( Zonemaster::Engine::Util::should_run_test( q{nameserver06} ) ) {
         push @results, $class->nameserver06( $zone );
     }
-    if ( Zonemaster::Engine->config->should_run( 'nameserver07' ) ) {
+    if ( Zonemaster::Engine::Util::should_run_test( q{nameserver07} ) ) {
         push @results, $class->nameserver07( $zone );
     }
-    if ( Zonemaster::Engine->config->should_run( 'nameserver08' ) ) {
+    if ( Zonemaster::Engine::Util::should_run_test( q{nameserver08} ) ) {
         push @results, $class->nameserver08( $zone );
     }
-    if ( Zonemaster::Engine->config->should_run( 'nameserver09' ) ) {
+    if ( Zonemaster::Engine::Util::should_run_test( q{nameserver09} ) ) {
         push @results, $class->nameserver09( $zone );
     }
+    if ( Zonemaster::Engine::Util::should_run_test( q{nameserver10} ) ) {
+        push @results, $class->nameserver10( $zone );
+    }
+    if ( Zonemaster::Engine::Util::should_run_test( q{nameserver11} ) ) {
+        push @results, $class->nameserver11( $zone );
+    }
+    if ( Zonemaster::Engine::Util::should_run_test( q{nameserver12} ) ) {
+        push @results, $class->nameserver12( $zone );
+    }
+    if ( Zonemaster::Engine::Util::should_run_test( q{nameserver13} ) ) {
+        push @results, $class->nameserver13( $zone );
+    }
 
     return @results;
 } ## end sub all
@@ -65,14 +84,18 @@ sub metadata {
             qw(
               IS_A_RECURSOR
               NO_RECURSOR
-              RECURSIVITY_UNDEF
+              NO_RESPONSE
               )
         ],
         nameserver02 => [
             qw(
-              EDNS0_BAD_QUERY
-              EDNS0_BAD_ANSWER
+              BREAKS_ON_EDNS
+              EDNS_RESPONSE_WITHOUT_EDNS
+              EDNS_VERSION_ERROR
               EDNS0_SUPPORT
+              NO_EDNS_SUPPORT
+              NO_RESPONSE
+              NS_ERROR
               )
         ],
         nameserver03 => [
@@ -91,6 +114,7 @@ sub metadata {
             qw(
               QUERY_DROPPED
               ANSWER_BAD_RCODE
+              AAAA_WELL_PROCESSED
               IPV4_DISABLED
               IPV6_DISABLED
               )
@@ -126,49 +150,88 @@ sub metadata {
               CASE_QUERIES_RESULTS_DIFFER
               )
         ],
+        nameserver10 => [
+            qw(
+              NO_RESPONSE
+              NO_EDNS_SUPPORT
+              UNSUPPORTED_EDNS_VER
+              NS_ERROR
+              )
+        ],
+        nameserver11 => [
+            qw(
+              NO_RESPONSE
+              NO_EDNS_SUPPORT
+              UNKNOWN_OPTION_CODE
+              NS_ERROR
+              )
+        ],
+        nameserver12 => [
+            qw(
+              NO_RESPONSE
+              NO_EDNS_SUPPORT
+              Z_FLAGS_NOTCLEAR
+              NS_ERROR
+              )
+        ],
+        nameserver13 => [
+            qw(
+              NO_RESPONSE
+              NO_EDNS_SUPPORT
+              NS_ERROR
+              MISSING_OPT_IN_TRUNCATED
+              )
+        ],
     };
 } ## end sub metadata
 
 sub translation {
     return {
-        'AAAA_WELL_PROCESSED' => 'The following nameservers answer AAAA queries without problems : {names}.',
-        'EDNS0_BAD_QUERY'     => 'Nameserver {ns}/{address} does not support EDNS0 (replies with FORMERR).',
-        'DIFFERENT_SOURCE_IP' =>
-          'Nameserver {ns}/{address} replies on a SOA query with a different source address ({source}).',
-        'SAME_SOURCE_IP'      => 'All nameservers reply with same IP used to query them.',
-        'AXFR_AVAILABLE'      => 'Nameserver {ns}/{address} allow zone transfer using AXFR.',
-        'AXFR_FAILURE'        => 'AXFR not available on nameserver {ns}/{address}.',
-        'QUERY_DROPPED'       => 'Nameserver {ns}/{address} dropped AAAA query.',
-        'IS_A_RECURSOR'       => 'Nameserver {ns}/{address} is a recursor.',
-        'NO_RECURSOR'         => 'None of the following nameservers is a recursor : {names}.',
-        'RECURSIVITY_UNDEF'   => 'Can not determine nameservers recursivity.',
-        'ANSWER_BAD_RCODE'    => 'Nameserver {ns}/{address} answered AAAA query with an unexpected rcode ({rcode}).',
-        'EDNS0_BAD_ANSWER'    => 'Nameserver {ns}/{address} does not support EDNS0 (OPT not set in reply).',
-        'EDNS0_SUPPORT'       => 'The following nameservers support EDNS0 : {names}.',
-        'CAN_NOT_BE_RESOLVED' => 'The following nameservers failed to resolve to an IP address : {names}.',
-        'CAN_BE_RESOLVED'     => 'All nameservers succeeded to resolve to an IP address.',
-        'NO_RESOLUTION'       => 'No nameservers succeeded to resolve to an IP address.',
-        'IPV4_DISABLED'       => 'IPv4 is disabled, not sending "{rrtype}" query to {ns}/{address}.',
-        'IPV6_DISABLED'       => 'IPv6 is disabled, not sending "{rrtype}" query to {ns}/{address}.',
-        'UPWARD_REFERRAL_IRRELEVANT' => 'Upward referral tests skipped for root zone.',
-        'UPWARD_REFERRAL'            => 'Nameserver {ns}/{address} returns an upward referral.',
-        'NO_UPWARD_REFERRAL'         => 'None of the following nameservers returns an upward referral : {names}.',
-        'QNAME_CASE_SENSITIVE'       => 'Nameserver {ns}/{address} preserves original case of queried names.',
-        'QNAME_CASE_INSENSITIVE'     => 'Nameserver {ns}/{address} does not preserve original case of queried names.',
-        'CASE_QUERY_SAME_ANSWER' =>
-          'When asked for {type} records on "{query1}" and "{query2}", nameserver {ns}/{address} returns same answers.',
-        'CASE_QUERY_DIFFERENT_ANSWER' =>
-'When asked for {type} records on "{query1}" and "{query2}", nameserver {ns}/{address} returns different answers.',
-        'CASE_QUERY_SAME_RC' =>
-'When asked for {type} records on "{query1}" and "{query2}", nameserver {ns}/{address} returns same RCODE "{rcode}".',
-        'CASE_QUERY_DIFFERENT_RC' =>
-'When asked for {type} records on "{query1}" and "{query2}", nameserver {ns}/{address} returns different RCODE ("{rcode1}" vs "{rcode2}").',
-        'CASE_QUERY_NO_ANSWER' =>
-          'When asked for {type} records on "{query}", nameserver {ns}/{address} returns nothing.',
-        'CASE_QUERIES_RESULTS_OK' =>
-          'When asked for {type} records on "{query}" with different cases, all servers reply consistently.',
-        'CASE_QUERIES_RESULTS_DIFFER' =>
-          'When asked for {type} records on "{query}" with different cases, all servers do not reply consistently.',
+        AAAA_WELL_PROCESSED => 'The following nameservers answer AAAA queries without problems : {names}.',
+        ANSWER_BAD_RCODE    => 'Nameserver {ns}/{address} answered AAAA query with an unexpected rcode ({rcode}).',
+        AXFR_AVAILABLE      => 'Nameserver {ns}/{address} allow zone transfer using AXFR.',
+        AXFR_FAILURE        => 'AXFR not available on nameserver {ns}/{address}.',
+        BREAKS_ON_EDNS      => 'No response from {ns}/{address} when EDNS is used in query asking for {dname}.',
+        CAN_BE_RESOLVED     => 'All nameservers succeeded to resolve to an IP address.',
+        CAN_NOT_BE_RESOLVED => 'The following nameservers failed to resolve to an IP address : {names}.',
+        CASE_QUERIES_RESULTS_DIFFER => 'When asked for {type} records on "{query}" with different cases, '
+          . 'all servers do not reply consistently.',
+        CASE_QUERIES_RESULTS_OK => 'When asked for {type} records on "{query}" with different cases, '
+          . 'all servers reply consistently.',
+        CASE_QUERY_DIFFERENT_ANSWER => 'When asked for {type} records on "{query1}" and "{query2}", '
+          . 'nameserver {ns}/{address} returns different answers.',
+        CASE_QUERY_DIFFERENT_RC => 'When asked for {type} records on "{query1}" and "{query2}", '
+          . 'nameserver {ns}/{address} returns different RCODE ("{rcode1}" vs "{rcode2}").',
+        CASE_QUERY_NO_ANSWER => 'When asked for {type} records on "{query}", '
+          . 'nameserver {ns}/{address} returns nothing.',
+        CASE_QUERY_SAME_ANSWER => 'When asked for {type} records on "{query1}" and "{query2}", '
+          . 'nameserver {ns}/{address} returns same answers.',
+        CASE_QUERY_SAME_RC => 'When asked for {type} records on "{query1}" and "{query2}", '
+          . 'nameserver {ns}/{address} returns same RCODE "{rcode}".',
+        DIFFERENT_SOURCE_IP => 'Nameserver {ns}/{address} replies on a SOA query with a different source address '
+          . '({source}).',
+        EDNS_RESPONSE_WITHOUT_EDNS => 'Response without EDNS from {ns}/{address} on query with EDNS0 asking for {dname}.',
+        EDNS_VERSION_ERROR         => 'Incorrect version of EDNS (expected 0) in response from {ns}/{address} on query with EDNS (version 0) asking for {dname}.',
+        EDNS0_SUPPORT              => 'The following nameservers support EDNS0 : {names}.',
+        IPV4_DISABLED              => 'IPv4 is disabled, not sending "{rrtype}" query to {ns}/{address}.',
+        IPV6_DISABLED              => 'IPv6 is disabled, not sending "{rrtype}" query to {ns}/{address}.',
+        IS_A_RECURSOR              => 'Nameserver {ns}/{address} is a recursor.',
+        MISSING_OPT_IN_TRUNCATED   => 'Nameserver {ns}/{address} replies on an EDNS query with a truncated response without EDNS.',
+        NO_EDNS_SUPPORT            => 'Nameserver {ns}/{address} does not support EDNS0 (replies with FORMERR).',
+        NO_RECURSOR                => 'Nameserver {ns}/{address} is not a recursor.',
+        NO_RESOLUTION              => 'No nameservers succeeded to resolve to an IP address.',
+        NO_RESPONSE                => 'No response from {ns}/{address} asking for {dname}.',
+        NO_UPWARD_REFERRAL         => 'None of the following nameservers returns an upward referral : {names}.',
+        NS_ERROR                   => 'Erroneous response from nameserver {ns}/{address}.',
+        QNAME_CASE_INSENSITIVE     => 'Nameserver {ns}/{address} does not preserve original case of queried names.',
+        QNAME_CASE_SENSITIVE       => 'Nameserver {ns}/{address} preserves original case of queried names.',
+        QUERY_DROPPED              => 'Nameserver {ns}/{address} dropped AAAA query.',
+        SAME_SOURCE_IP             => 'All nameservers reply with same IP used to query them.',
+        UNKNOWN_OPTION_CODE        => 'Nameserver {ns}/{address} responds with an unknown ENDS OPTION-CODE.',
+        UNSUPPORTED_EDNS_VER       => 'Nameserver {ns}/{address} accepts an unsupported EDNS version.',
+        UPWARD_REFERRAL            => 'Nameserver {ns}/{address} returns an upward referral.',
+        UPWARD_REFERRAL_IRRELEVANT => 'Upward referral tests skipped for root zone.',
+        Z_FLAGS_NOTCLEAR           => 'Nameserver {ns}/{address} has one or more unknown EDNS Z flag bits set.',
     };
 } ## end sub translation
 
@@ -178,139 +241,69 @@ sub version {
 
 sub nameserver01 {
     my ( $class, $zone ) = @_;
-    my $nonexistent_name = q{xx--domain-cannot-exist.xx--illegal-syntax-tld};
-    my $unlikely_label   = q{xx--domain-should-not-exist};
-    my @existing_tld     = qw{fr re pm tf yt wf si};
     my @results;
-    my %ips;
-    my %nsnames;
-    my %is_not_recursor = ();
 
-    foreach
-      my $local_ns ( @{ Zonemaster::Engine::TestMethods->method4( $zone ) }, @{ Zonemaster::Engine::TestMethods->method5( $zone ) } )
+    my @nss;
     {
-
-        next if ( not Zonemaster::Engine->config->ipv6_ok and $local_ns->address->version == $IP_VERSION_6 );
-
-        next if ( not Zonemaster::Engine->config->ipv4_ok and $local_ns->address->version == $IP_VERSION_4 );
-
-        next if $ips{ $local_ns->address->short };
-
-        my $p = $local_ns->query( $nonexistent_name, q{SOA}, { recurse => 1 } );
-
-        if ( $p ) {
-            if ( $p->rcode eq q{REFUSED} ) {
-                $is_not_recursor{ $local_ns->address->short }++;
-            }
-            elsif ( $p->rcode eq q{SERVFAIL} ) {
-                $is_not_recursor{ $local_ns->address->short }++;
-            }
-            elsif ( $p->rcode eq q{NXDOMAIN} and not $p->aa ) {
-                push @results,
-                  info(
-                    IS_A_RECURSOR => {
-                        ns      => $local_ns->name,
-                        address => $local_ns->address->short,
-                        dname   => $nonexistent_name,
-                    }
-                  );
-            }
-            elsif ( $p->is_redirect and not $p->aa ) {
-                $is_not_recursor{ $local_ns->address->short }++;
-            }
-            elsif ( not $p->is_redirect and not $p->aa and not $p->answer and $p->rcode eq q{NOERROR} ) {
-                $is_not_recursor{ $local_ns->address->short }++;
-            }
-            $nsnames{ $local_ns->name }++;
-            $ips{ $local_ns->address->short }++;
-        }
-
-    } ## end foreach my $local_ns ( @{ Zonemaster::Engine::TestMethods...})
-
-    my $ips_string = join '#', sort keys %ips;
-    my $is_not_recursor_string = join '#', sort keys %is_not_recursor;
-    if ( $ips_string and $ips_string eq $is_not_recursor_string ) {
-        push @results,
-          info(
-            NO_RECURSOR => {
-                names => join( q{,}, sort keys %nsnames ),
-            }
-          );
+        my %nss = map { $_->string => $_ }
+          @{ Zonemaster::Engine::TestMethods->method4( $zone ) },
+          @{ Zonemaster::Engine::TestMethods->method5( $zone ) };
+        @nss = values %nss;
     }
-    elsif ( not grep { $_->tag eq q{IS_A_RECURSOR} } @results ) {
-        foreach my $tld ( @existing_tld ) {
 
-            next if $tld eq $zone->name;
-            my $checking_name = $unlikely_label . q{.} . $tld;
-            %is_not_recursor = ();
-            %ips = ();
+    if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv6}) ) {
+        @nss = grep { $_->address->version != $IP_VERSION_6 } @nss;
+    }
+    if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv4}) ) {
+        @nss = grep { $_->address->version != $IP_VERSION_4 } @nss;
+    }
 
-            foreach my $local_ns ( @{ Zonemaster::Engine::TestMethods->method4( $zone ) }, @{ Zonemaster::Engine::TestMethods->method5( $zone ) } ) {
+    for my $ns ( @nss ) {
+        my %ns_args = (
+            ns      => $ns->name->string,
+            address => $ns->address->short,
+        );
 
-                next if ( not Zonemaster::Engine->config->ipv6_ok and $local_ns->address->version == $IP_VERSION_6 );
+        my $response_count = 0;
+        my $nxdomain_count = 0;
+        my $is_no_recursor = 1;
+        my $has_seen_ra    = 0;
+        for my $nonexistent_name ( @NONEXISTENT_NAMES ) {
 
-                next if ( not Zonemaster::Engine->config->ipv4_ok and $local_ns->address->version == $IP_VERSION_4 );
+            my $p = $ns->query( $nonexistent_name, q{A}, { blacklisting_disabled => 1 } );
+            if ( !$p ) {
+                my %name_args = (
+                    dname => $nonexistent_name,
+                    %ns_args,
+                );
+                push @results, info( NO_RESPONSE => \%name_args );
+                $is_no_recursor = 0;
+            }
+            else {
+                $response_count++;
 
-                next if $ips{ $local_ns->address->short };
+                if ( $p->ra ) {
+                    $has_seen_ra = 1;
+                }
 
-                next if $is_not_recursor{ $local_ns->address->short };
-
-                my $p = $local_ns->query( $checking_name, q{SOA}, { recurse => 1 } );
-
-                if ( $p ) {
-                    if ( $p->rcode eq q{REFUSED} ) {
-                        $is_not_recursor{ $local_ns->address->short }++;
-                    }
-                    elsif ( $p->rcode eq q{SERVFAIL} ) {
-                        $is_not_recursor{ $local_ns->address->short }++;
-                    }
-                    elsif ( $p->rcode eq q{NXDOMAIN} and not $p->aa ) {
-                        push @results,
-                          info(
-                            IS_A_RECURSOR => {
-                                ns      => $local_ns->name,
-                                address => $local_ns->address->short,
-                                dname   => $checking_name,
-                            }
-                          );
-                    }
-                    elsif ( $p->is_redirect and not $p->aa ) {
-                        $is_not_recursor{ $local_ns->address->short }++;
-                    }
-                    elsif ( not $p->is_redirect and not $p->aa and not $p->answer and $p->rcode eq q{NOERROR} ) {
-                        $is_not_recursor{ $local_ns->address->short }++;
-                    }
-                    $nsnames{ $local_ns->name }++;
-                    $ips{ $local_ns->address->short }++;
+                if ( $p->rcode eq q{NXDOMAIN} ) {
+                    $nxdomain_count++;
                 }
             }
+        } ## end for my $nonexistent_name...
 
-            my $ips_string = join '#', sort keys %ips;
-            my $is_not_recursor_string = join '#', sort keys %is_not_recursor;
-            if ( $ips_string and $ips_string eq $is_not_recursor_string ) {
-                push @results,
-                  info(
-                    NO_RECURSOR => {
-                        names => join( q{,}, sort keys %nsnames ),
-                    }
-                  );
-                last;
-            }
-
-            if ( grep { $_->tag eq q{IS_A_RECURSOR} } @results ) {
-                last;
-            }
+        if ( $has_seen_ra || ( $response_count > 0 && $nxdomain_count == $response_count ) ) {
+            push @results, info( IS_A_RECURSOR => \%ns_args );
+            $is_no_recursor = 0;
         }
 
-        if ( not grep { $_->tag eq q{IS_A_RECURSOR} } @results and not grep { $_->tag eq q{NO_RECURSOR} } @results ) {
-            push @results,
-              info(
-                RECURSIVITY_UNDEF => {}
-              );
+        if ( $is_no_recursor ) {
+            push @results, info( NO_RECURSOR => \%ns_args );
         }
-    }
+    } ## end for my $ns ( @nss )
 
     return @results;
+
 } ## end sub nameserver01
 
 sub nameserver02 {
@@ -321,36 +314,80 @@ sub nameserver02 {
     foreach
       my $local_ns ( @{ Zonemaster::Engine::TestMethods->method4( $zone ) }, @{ Zonemaster::Engine::TestMethods->method5( $zone ) } )
     {
+        next if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv6}) and $local_ns->address->version == $IP_VERSION_6 );
 
-        next if ( not Zonemaster::Engine->config->ipv6_ok and $local_ns->address->version == $IP_VERSION_6 );
-
-        next if ( not Zonemaster::Engine->config->ipv4_ok and $local_ns->address->version == $IP_VERSION_4 );
+        next if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv4}) and $local_ns->address->version == $IP_VERSION_4 );
 
         next if $nsnames_and_ip{ $local_ns->name->string . q{/} . $local_ns->address->short };
 
         my $p = $local_ns->query( $zone->name, q{SOA}, { edns_size => 512 } );
         if ( $p ) {
-            if ( $p->rcode eq q{FORMERR} ) {
+            if ( $p->rcode eq q{FORMERR} and not $p->has_edns) {
                 push @results,
                   info(
-                    EDNS0_BAD_QUERY => {
+                    NO_EDNS_SUPPORT => {
                         ns      => $local_ns->name,
                         address => $local_ns->address->short,
                     }
                   );
             }
-            else {
-                if ( not $p->has_edns ) {
-                    push @results,
-                      info(
-                        EDNS0_BAD_ANSWER => {
-                            ns      => $local_ns->name,
-                            address => $local_ns->address->short,
-                        }
-                      );
-                }
+            elsif ( $p->rcode eq q{NOERROR} and not $p->edns_rcode and $p->get_records( q{SOA}, q{answer} ) and $p->edns_version == 0 ) {
+                $nsnames_and_ip{ $local_ns->name->string . q{/} . $local_ns->address->short }++;
+                next;
             }
-        } ## end if ( $p )
+            elsif ( $p->rcode eq q{NOERROR} and not $p->has_edns ) {
+                push @results,
+                  info(
+                    EDNS_RESPONSE_WITHOUT_EDNS => {
+                        ns      => $local_ns->name,
+                        address => $local_ns->address->short,
+                        dname   => $zone->name,
+                    }
+                  );
+            }
+            elsif ( $p->rcode eq q{NOERROR} and $p->has_edns and $p->edns_version != 0 ) {
+                push @results,
+                  info(
+                    EDNS_VERSION_ERROR => {
+                        ns      => $local_ns->name,
+                        address => $local_ns->address->short,
+                        dname   => $zone->name,
+                    }
+                  );
+            }
+            else {
+                push @results,
+                  info(
+                    NS_ERROR => {
+                        ns      => $local_ns->name,
+                        address => $local_ns->address->short,
+                    }
+                  );
+            }
+        }
+        else {
+            my $p2 = $local_ns->query( $zone->name, q{SOA} );
+            if ( $p2 ) {
+                push @results,
+                  info(
+                    BREAKS_ON_EDNS => {
+                        ns      => $local_ns->name,
+                        address => $local_ns->address->short,
+                        dname   => $zone->name,
+                    }
+                  );
+            }
+	    else {
+                push @results,
+                  info(
+                    NO_RESPONSE => {
+                        ns      => $local_ns->name,
+                        address => $local_ns->address->short,
+                        dname   => $zone->name,
+                    }
+                  );
+            }
+        }
 
         $nsnames_and_ip{ $local_ns->name->string . q{/} . $local_ns->address->short }++;
     } ## end foreach my $local_ns ( @{ Zonemaster::Engine::TestMethods...})
@@ -376,9 +413,9 @@ sub nameserver03 {
       my $local_ns ( @{ Zonemaster::Engine::TestMethods->method4( $zone ) }, @{ Zonemaster::Engine::TestMethods->method5( $zone ) } )
     {
 
-        next if ( not Zonemaster::Engine->config->ipv6_ok and $local_ns->address->version == $IP_VERSION_6 );
+        next if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv6}) and $local_ns->address->version == $IP_VERSION_6 );
 
-        next if ( not Zonemaster::Engine->config->ipv4_ok and $local_ns->address->version == $IP_VERSION_4 );
+        next if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv4}) and $local_ns->address->version == $IP_VERSION_4 );
 
         next if $nsnames_and_ip{ $local_ns->name->string . q{/} . $local_ns->address->short };
 
@@ -421,9 +458,9 @@ sub nameserver04 {
       my $local_ns ( @{ Zonemaster::Engine::TestMethods->method4( $zone ) }, @{ Zonemaster::Engine::TestMethods->method5( $zone ) } )
     {
 
-        next if ( not Zonemaster::Engine->config->ipv6_ok and $local_ns->address->version == $IP_VERSION_6 );
+        next if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv6}) and $local_ns->address->version == $IP_VERSION_6 );
 
-        next if ( not Zonemaster::Engine->config->ipv4_ok and $local_ns->address->version == $IP_VERSION_4 );
+        next if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv4}) and $local_ns->address->version == $IP_VERSION_4 );
 
         next if $nsnames_and_ip{ $local_ns->name->string . q{/} . $local_ns->address->short };
 
@@ -467,7 +504,7 @@ sub nameserver05 {
 
         next if $nsnames_and_ip{ $local_ns->name->string . q{/} . $local_ns->address->short };
 
-        if ( not Zonemaster::Engine->config->ipv6_ok and $local_ns->address->version == $IP_VERSION_6 ) {
+        if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv6}) and $local_ns->address->version == $IP_VERSION_6 ) {
             push @results,
               info(
                 IPV6_DISABLED => {
@@ -479,7 +516,7 @@ sub nameserver05 {
             next;
         }
 
-        if ( not Zonemaster::Engine->config->ipv4_ok and $local_ns->address->version == $IP_VERSION_4 ) {
+        if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv4}) and $local_ns->address->version == $IP_VERSION_4 ) {
             push @results,
               info(
                 IPV4_DISABLED => {
@@ -588,13 +625,13 @@ sub nameserver07 {
         foreach my $local_ns ( @{ Zonemaster::Engine::TestMethods->method4( $zone ) },
             @{ Zonemaster::Engine::TestMethods->method5( $zone ) } )
         {
-            next if ( not Zonemaster::Engine->config->ipv6_ok and $local_ns->address->version == $IP_VERSION_6 );
+            next if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv6}) and $local_ns->address->version == $IP_VERSION_6 );
 
-            next if ( not Zonemaster::Engine->config->ipv4_ok and $local_ns->address->version == $IP_VERSION_4 );
+            next if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv4}) and $local_ns->address->version == $IP_VERSION_4 );
 
             next if $nsnames_and_ip{ $local_ns->name->string . q{/} . $local_ns->address->short };
 
-            my $p = $local_ns->query( q{.}, q{NS} );
+	    my $p = $local_ns->query( q{.}, q{NS}, { blacklisting_disabled => 1 } );
             if ( $p ) {
                 my @ns = $p->get_records( q{NS}, q{authority} );
 
@@ -641,9 +678,9 @@ sub nameserver08 {
     foreach
       my $local_ns ( @{ Zonemaster::Engine::TestMethods->method4( $zone ) }, @{ Zonemaster::Engine::TestMethods->method5( $zone ) } )
     {
-        next if ( not Zonemaster::Engine->config->ipv6_ok and $local_ns->address->version == $IP_VERSION_6 );
+        next if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv6}) and $local_ns->address->version == $IP_VERSION_6 );
 
-        next if ( not Zonemaster::Engine->config->ipv4_ok and $local_ns->address->version == $IP_VERSION_4 );
+        next if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv4}) and $local_ns->address->version == $IP_VERSION_4 );
 
         next if $nsnames_and_ip{ $local_ns->name->string . q{/} . $local_ns->address->short };
 
@@ -702,9 +739,9 @@ sub nameserver09 {
     foreach
       my $local_ns ( @{ Zonemaster::Engine::TestMethods->method4( $zone ) }, @{ Zonemaster::Engine::TestMethods->method5( $zone ) } )
     {
-        next if ( not Zonemaster::Engine->config->ipv6_ok and $local_ns->address->version == $IP_VERSION_6 );
+        next if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv6}) and $local_ns->address->version == $IP_VERSION_6 );
 
-        next if ( not Zonemaster::Engine->config->ipv4_ok and $local_ns->address->version == $IP_VERSION_4 );
+        next if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv4}) and $local_ns->address->version == $IP_VERSION_4 );
 
         next if $nsnames_and_ip{ $local_ns->name->string . q{/} . $local_ns->address->short };
 
@@ -822,6 +859,286 @@ sub nameserver09 {
     return @results;
 } ## end sub nameserver09
 
+sub nameserver10 {
+    my ( $class, $zone ) = @_;
+    my @results;
+
+    my @nss;
+    {
+        my %nss = map { $_->string => $_ }
+          @{ Zonemaster::Engine::TestMethods->method4( $zone ) },
+          @{ Zonemaster::Engine::TestMethods->method5( $zone ) };
+        @nss = values %nss;
+    }
+
+    if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv6}) ) {
+        @nss = grep { $_->address->version != $IP_VERSION_6 } @nss;
+    }
+    if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv4}) ) {
+        @nss = grep { $_->address->version != $IP_VERSION_4 } @nss;
+    }
+
+    for my $ns ( @nss ) {
+	my $p = $ns->query( $zone->name, q{SOA}, { edns_details => { version => 1 } } );
+        if ( $p ) {
+            if ( $p->rcode eq q{FORMERR} and not $p->edns_rcode ) {
+                push @results,
+                  info(
+                    NO_EDNS_SUPPORT => {
+                        ns      => $ns->name,
+                        address => $ns->address->short,
+                    }
+                  );
+            }
+            elsif ( $p->rcode eq q{NOERROR} and not $p->edns_rcode ) {
+                push @results,
+                  info(
+                    UNSUPPORTED_EDNS_VER => {
+                        ns      => $ns->name,
+                        address => $ns->address->short,
+                    }
+                  );
+            }
+            elsif ( ($p->rcode eq q{NOERROR} and $p->edns_rcode == 1) and $p->edns_version == 0 and not scalar $p->answer) {
+                next;
+            }
+            else {
+                push @results,
+                  info(
+                    NS_ERROR => {
+                        ns      => $ns->name,
+                        address => $ns->address->short,
+                    }
+                  );
+            }
+        }
+	else {
+            push @results,
+              info(
+                NO_RESPONSE => {
+                    ns      => $ns->name,
+                    address => $ns->address->short,
+                    dname   => $zone->name,
+                }
+              );
+        }
+    }
+
+    return @results;
+} ## end sub nameserver10
+
+sub nameserver11 {
+    my ( $class, $zone ) = @_;
+    my @results;
+
+    my @nss;
+    {
+        my %nss = map { $_->string => $_ }
+          @{ Zonemaster::Engine::TestMethods->method4( $zone ) },
+          @{ Zonemaster::Engine::TestMethods->method5( $zone ) };
+        @nss = values %nss;
+    }
+
+    if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv6}) ) {
+        @nss = grep { $_->address->version != $IP_VERSION_6 } @nss;
+    }
+    if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv4}) ) {
+        @nss = grep { $_->address->version != $IP_VERSION_4 } @nss;
+    }
+
+    # Choose an unassigned EDNS0 Option Codes
+    # values 15-26945 are Unassigned. Let's say we use 137 ???
+    my $opt_code = 137;
+    my $opt_data = q{};
+    my $opt_length = length($opt_data);
+    my $rdata = $opt_code*65536 + $opt_length;
+
+    for my $ns ( @nss ) {
+        my $p = $ns->query( $zone->name, q{SOA}, { edns_details => { data => $rdata } } );
+        if ( $p ) {
+            if ( $p->rcode eq q{FORMERR} and not $p->edns_rcode ) {
+                push @results,
+                  info(
+                    NO_EDNS_SUPPORT => {
+                        ns      => $ns->name,
+                        address => $ns->address->short,
+                    }
+                  );
+            }
+            elsif ( defined $p->edns_data ) {
+                push @results,
+                  info(
+                    UNKNOWN_OPTION_CODE => {
+                        ns      => $ns->name,
+                        address => $ns->address->short,
+                    }
+                  );
+            }
+            elsif ( $p->rcode eq q{NOERROR} and not $p->edns_rcode and $p->edns_version == 0 and not defined $p->edns_data and $p->get_records( q{SOA}, q{answer} ) ) {
+                next;
+            }
+            else {
+                push @results,
+                  info(
+                    NS_ERROR => {
+                        ns      => $ns->name,
+                        address => $ns->address->short,
+                    }
+                  );
+            }
+        }
+        else {
+            push @results,
+              info(
+                NO_RESPONSE => {
+                    ns      => $ns->name,
+                    address => $ns->address->short,
+                    dname   => $zone->name,
+                }
+              );
+        }
+
+    }
+
+    return @results;
+} ## end sub nameserver11
+
+sub nameserver12 {
+    my ( $class, $zone ) = @_;
+    my @results;
+
+    my @nss;
+    {
+        my %nss = map { $_->string => $_ }
+          @{ Zonemaster::Engine::TestMethods->method4( $zone ) },
+          @{ Zonemaster::Engine::TestMethods->method5( $zone ) };
+        @nss = values %nss;
+    }
+
+    if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv6}) ) {
+        @nss = grep { $_->address->version != $IP_VERSION_6 } @nss;
+    }
+    if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv4}) ) {
+        @nss = grep { $_->address->version != $IP_VERSION_4 } @nss;
+    }
+
+    for my $ns ( @nss ) {
+        my $p = $ns->query( $zone->name, q{SOA}, { edns_details => { z => 3 } } );
+        if ( $p ) {
+            if ( $p->rcode eq q{FORMERR} and not $p->edns_rcode ) {
+                push @results,
+                  info(
+                    NO_EDNS_SUPPORT => {
+                        ns      => $ns->name,
+                        address => $ns->address->short,
+                    }
+                  );
+            }
+            elsif ( $p->edns_z ) {
+                push @results,
+                  info(
+                    Z_FLAGS_NOTCLEAR => {
+                        ns      => $ns->name,
+                        address => $ns->address->short,
+                    }
+                  );
+            }
+            elsif ( $p->rcode eq q{NOERROR} and not $p->edns_rcode and $p->edns_version == 0 and $p->edns_z == 0 and $p->get_records( q{SOA}, q{answer} ) ) {
+                next;
+            }
+            else {
+                push @results,
+                  info(
+                    NS_ERROR => {
+                        ns      => $ns->name,
+                        address => $ns->address->short,
+                    }
+                  );
+            }
+        }
+        else {
+            push @results,
+              info(
+                NO_RESPONSE => {
+                    ns      => $ns->name,
+                    address => $ns->address->short,
+                    dname   => $zone->name,
+                }
+              );
+        }
+    }
+
+    return @results;
+} ## end sub nameserver12
+
+sub nameserver13 {
+    my ( $class, $zone ) = @_;
+    my @results;
+
+    my @nss;
+    {
+        my %nss = map { $_->string => $_ }
+          @{ Zonemaster::Engine::TestMethods->method4( $zone ) },
+          @{ Zonemaster::Engine::TestMethods->method5( $zone ) };
+        @nss = values %nss;
+    }
+
+    if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv6}) ) {
+        @nss = grep { $_->address->version != $IP_VERSION_6 } @nss;
+    }
+    if ( not Zonemaster::Engine::Profile->effective->get(q{net.ipv4}) ) {
+        @nss = grep { $_->address->version != $IP_VERSION_4 } @nss;
+    }
+
+    for my $ns ( @nss ) {
+        my $p = $ns->query( $zone->name, q{SOA}, { usevc => 0, fallback => 0, edns_details => { do => 1, udp_size => 512  } } );
+        if ( $p ) {
+            if ( $p->rcode eq q{FORMERR} and not $p->edns_rcode ) {
+                push @results,
+                  info(
+                    NO_EDNS_SUPPORT => {
+                        ns      => $ns->name,
+                        address => $ns->address->short,
+                    }
+                  );
+            }
+            elsif ( $p->tc and not $p->has_edns ) {
+                push @results,
+                  info(
+                    MISSING_OPT_IN_TRUNCATED => {
+                        ns      => $ns->name,
+                        address => $ns->address->short,
+                    }
+                  );
+            }
+            elsif ( $p->rcode eq q{NOERROR} and not $p->edns_rcode and $p->edns_version == 0 ) {
+                next;
+            }
+            else {
+                push @results,
+                  info(
+                    NS_ERROR => {
+                        ns      => $ns->name,
+                        address => $ns->address->short,
+                    }
+                  );
+            }
+        }
+        else {
+            push @results,
+              info(
+                NO_RESPONSE => {
+                    ns      => $ns->name,
+                    address => $ns->address->short,
+                    dname   => $zone->name,
+                }
+              );
+        }
+    }
+
+    return @results;
+} ## end sub nameserver13
+
 1;
 
 =head1 NAME
@@ -842,7 +1159,7 @@ Runs the default set of tests and returns a list of log entries made by the test
 
 =item translation()
 
-Returns a refernce to a hash with translation data. Used by the builtin translation system.
+Returns a reference to a hash with translation data. Used by the builtin translation system.
 
 =item metadata()
 
@@ -895,6 +1212,22 @@ Check whether authoritative name servers responses match the case of every lette
 
 Check whether authoritative name servers return same results for equivalent names with different cases in the request.
 
+=item nameserver10($zone)
+
+WIP
+
+=item nameserver11($zone)
+
+WIP
+
+=item nameserver12($zone)
+
+WIP
+
+=item nameserver13($zone)
+
+WIP
+
 =back
 
 =cut
